@@ -21,7 +21,7 @@ let ml_automata = ref false
 let source_name = ref None
 let output_name = ref None
 
-let usage = "usage: ocamllex [options] sourcefile"
+let usage = "Usage: ocamllex [options] sourcefile\nOptions are:"
 
 let print_version_string () =
   print_string "The OCaml lexer generator, version ";
@@ -30,8 +30,7 @@ let print_version_string () =
 
 let print_version_num () =
   print_endline Sys.ocaml_version;
-  exit 0;
-;;
+  exit 0
 
 let specs =
   ["-ml", Arg.Set ml_automata,
@@ -88,6 +87,7 @@ let main () =
     close_out oc;
     Common.close_tracker tr;
   with exn ->
+    let bt = Printexc.get_raw_backtrace () in
     close_in ic;
     close_out oc;
     Common.close_tracker tr;
@@ -118,8 +118,8 @@ let main () =
           "File \"%s\":\ntransition table overflow, automaton is too big\n"
           source_name
     | _ ->
-        raise exn
+        Printexc.raise_with_backtrace exn bt
     end;
     exit 3
 
-let _ = (* Printexc.catch *) main (); exit 0
+let _ = main (); exit 0
