@@ -23,9 +23,15 @@ val cwd : Variables.t
 
 val commandline : Variables.t
 
+val dev_null : Variables.t
+
+val dst : Variables.t
+
 val exit_status : Variables.t
 
-val files : Variables.t
+val file : Variables.t
+
+val readonly_files : Variables.t
 
 val make : Variables.t
 
@@ -49,10 +55,14 @@ val skip_header_bytes : Variables.t
 
 val script : Variables.t
 
+val src : Variables.t
+
 val stdin : Variables.t
 val stdout : Variables.t
 val stderr : Variables.t
 
+val subdirectories : Variables.t
+
 val test_build_directory : Variables.t
 val test_build_directory_prefix : Variables.t
 
@@ -65,3 +75,5 @@ val test_pass : Variables.t
 val test_skip : Variables.t
 
 val test_fail : Variables.t
+
+val timeout : Variables.t
