@@ -120,9 +120,9 @@
                      | atomic_exchange(seq_cst)|
                      | fence(release)          |
    ------------------|-------------------------|--------------------------------
-   Non-atomic load   | atomic_load(relaxed)    | __tsan_read8()
+   Non-atomic load   | atomic_load(relaxed)    | __tsan_readN()
    ------------------|-------------------------|--------------------------------
-   Non-atomic store  | fence(acquire)          | __tsan_write8()
+   Non-atomic store  | fence(acquire)          | __tsan_writeN()
    (assignment, int) | atomic_store(release)   |
    ------------------|-------------------------|--------------------------------
    Non-atomic store  | fence(acquire)          | __tsan_write8()
@@ -133,7 +133,7 @@
    (initialization)  |                         |
    ------------------|-------------------------|--------------------------------
    Non-atomic store  | plain store             | __tsan_writeN()
-   (unaligned size)  |                         |
+   (unaligned)       |                         |
    ------------------|-------------------------|--------------------------------
 
    This mapping dictates which instrumentation calls are generated by the
@@ -159,7 +159,7 @@
      the FFI rules on the condition that the GC does not run between the
      allocation and the end of initialization) and a conflicting access is made
      from OCaml after publication to other threads. There should be no data
-     race thanks to data dependency (see [MMOC] coment in memory.c), but TSan
+     race thanks to data dependency (see [MMOC] comment in memory.c), but TSan
      does not take data dependencies into account.
    - A field is accessed from C with `Field`, or more generally using a
      `volatile value *` or a relaxed atomic access, and that field is modified
@@ -185,7 +185,7 @@
    More details and examples can be found in PR #12681.
 
    Our current make-do solution is that `__tsan_volatile_readN` performs a
-   dummy call to `__tsan_atomic64_load`, which is sufficient for TSan to view
+   dummy call to `__tsan_atomicNN_load`, which is sufficient for TSan to view
    them as relaxed loads; and `__tsan_volatile_writeN` performs a dummy
    fetch_add of zero. */
 
@@ -215,7 +215,7 @@ Caml_inline void caml_tsan_debug_log_pc(const char* msg, uintnat pc)
 void caml_tsan_exit_on_raise(uintnat pc, char* sp, char* trapsp)
 {
   caml_domain_state* domain_state = Caml_state;
-  caml_frame_descrs fds = caml_get_frame_descrs();
+  caml_frame_descrs* fds = caml_get_frame_descrs();
   uintnat next_pc = pc;
 
   /* iterate on each frame  */
@@ -233,7 +233,7 @@ void caml_tsan_exit_on_raise(uintnat pc, char* sp, char* trapsp)
     }
 
     caml_tsan_debug_log_pc("forced__tsan_func_exit for", pc);
-    __tsan_func_exit(NULL);
+    caml_tsan_func_exit();
     pc = next_pc;
   }
 }
@@ -248,7 +248,6 @@ void caml_tsan_exit_on_raise_c(char* limit)
 {
   unw_context_t uc;
   unw_cursor_t cursor;
-  unw_word_t sp;
 #ifdef TSAN_DEBUG
   unw_word_t prev_pc;
 #endif
@@ -261,12 +260,23 @@ void caml_tsan_exit_on_raise_c(char* limit)
   if (ret != 0)
     caml_fatal_error("unw_init_local failed with code %d", ret);
 
-  while (1) {
+  unw_word_t initial_sp;
+  ret = unw_get_reg(&cursor, UNW_REG_SP, &initial_sp);
+  if (ret != 0)
+    caml_fatal_error("unw_get_reg SP failed with code %d", ret);
+
+  /* Unwind each call in the stack fragment between `initial_sp` and `limit`. */
+  for(unw_word_t sp = initial_sp; initial_sp <= sp && (char*)sp < limit; ) {
+
 #ifdef TSAN_DEBUG
     if (unw_get_reg(&cursor, UNW_REG_IP, &prev_pc) < 0) {
       caml_fatal_error("unw_get_reg IP failed with code %d", ret);
     }
+
+    caml_tsan_debug_log_pc("forced__tsan_func_exit for", prev_pc);
 #endif
+    /* Still on the C stack, pop on the TSan shadow stack. */
+    caml_tsan_func_exit();
 
     ret = unw_step(&cursor);
     if (ret < 0) {
@@ -279,14 +289,6 @@ void caml_tsan_exit_on_raise_c(char* limit)
     ret = unw_get_reg(&cursor, UNW_REG_SP, &sp);
     if (ret != 0)
       caml_fatal_error("unw_get_reg SP failed with code %d", ret);
-#ifdef TSAN_DEBUG
-    caml_tsan_debug_log_pc("forced__tsan_func_exit for", prev_pc);
-#endif
-    __tsan_func_exit(NULL);
-
-    if ((char*)sp >= limit) {
-      break;
-    }
   }
 }
 
@@ -299,7 +301,7 @@ void caml_tsan_exit_on_raise_c(char* limit)
 void caml_tsan_exit_on_perform(uintnat pc, char* sp)
 {
   struct stack_info* stack = Caml_state->current_stack;
-  caml_frame_descrs fds = caml_get_frame_descrs();
+  caml_frame_descrs* fds = caml_get_frame_descrs();
   uintnat next_pc = pc;
 
   /* iterate on each frame  */
@@ -310,7 +312,7 @@ void caml_tsan_exit_on_perform(uintnat pc, char* sp)
     }
 
     caml_tsan_debug_log_pc("forced__tsan_func_exit for", pc);
-    __tsan_func_exit(NULL);
+    caml_tsan_func_exit();
 
     pc = next_pc;
   }
@@ -330,7 +332,7 @@ void caml_tsan_exit_on_perform(uintnat pc, char* sp)
 CAMLno_tsan void caml_tsan_entry_on_resume(uintnat pc, char* sp,
     struct stack_info const* stack)
 {
-  caml_frame_descrs fds = caml_get_frame_descrs();
+  caml_frame_descrs* fds = caml_get_frame_descrs();
   uintnat next_pc = pc;
 
   caml_next_frame_descriptor(fds, &next_pc, &sp, (struct stack_info*)stack);
@@ -346,24 +348,31 @@ CAMLno_tsan void caml_tsan_entry_on_resume(uintnat pc, char* sp,
 
   caml_tsan_entry_on_resume(next_pc, sp, stack);
   caml_tsan_debug_log_pc("forced__tsan_func_entry for", pc);
-  __tsan_func_entry((void*)next_pc);
+  caml_tsan_func_entry((void*)next_pc);
 }
 
 #endif // NATIVE_CODE
 
-
 #include "caml/mlvalues.h"
+
 #include <stdbool.h>
+#include <stdint.h>
+
+/*
+   Make TSan see aligned volatile reads, and word-aligned volatile writes as
+   relaxed atomic accesses.
+
+   Refer to the detailed comments at the beginning of this file.
+
+   [The `is_aligned' checks are probably unnecessary here as one would expect
+   unaligned accesses to cause __tsan_unaligned_volatile_* to be invoked, but
+   they don't hurt and we'd better be correct here.] */
 
 Caml_inline bool is_aligned(void *ptr, size_t byte_count)
 {
   return (uintptr_t)ptr % byte_count == 0;
 }
 
-#include <stdint.h>
-
-/* Make TSan see word-aligned volatile accesses as relaxed atomic accesses.
-   Refer to the detailed comments at the beginning of this file. */
 #define DEFINE_TSAN_VOLATILE_READ_WRITE(size, bitsize)                         \
                                                                                \
 extern void __tsan_read##size(void*);                                          \
@@ -375,7 +384,10 @@ extern uint##bitsize##_t __tsan_atomic##bitsize##_fetch_add(                   \
                                                                                \
 CAMLno_tsan void __tsan_volatile_read##size(void *ptr)                         \
 {                                                                              \
-  const bool is_atomic = size <= sizeof(long long) && is_aligned(ptr, 8);      \
+  /* read accesses are considered atomic if they are not larger than the       \
+     native register size, and the address is properly aligned for the         \
+     width of the access. */                                                   \
+  const bool is_atomic = size <= sizeof(void *) && is_aligned(ptr, size);      \
   if (is_atomic)                                                               \
     __tsan_atomic##bitsize##_load(ptr, memory_order_relaxed);                  \
   else                                                                         \
@@ -383,11 +395,14 @@ CAMLno_tsan void __tsan_volatile_read##size(void *ptr)                         \
 }                                                                              \
 CAMLno_tsan void __tsan_unaligned_volatile_read##size(void *ptr)               \
 {                                                                              \
-  __tsan_volatile_read##size(ptr);                                             \
+  __tsan_read##size(ptr);                                                      \
 }                                                                              \
 CAMLno_tsan void __tsan_volatile_write##size(void *ptr)                        \
 {                                                                              \
-  const bool is_atomic = size <= sizeof(long long) && is_aligned(ptr, 8);      \
+  /* write accesses are considered atomic only if they are the size of the     \
+     native register, and the address is properly aligned for the              \
+     width of the access. */                                                   \
+  const bool is_atomic = size == sizeof(void *) && is_aligned(ptr, size);      \
   if (is_atomic) {                                                             \
     /* Signal a relaxed atomic store to TSan. We don't have access to the      \
        actual value written so we do a fetch_add of 0 which has the effect of  \
@@ -398,7 +413,7 @@ CAMLno_tsan void __tsan_volatile_write##size(void *ptr)                        \
 }                                                                              \
 CAMLno_tsan void __tsan_unaligned_volatile_write##size(void *ptr)              \
 {                                                                              \
-  __tsan_volatile_write##size(ptr);                                            \
+  __tsan_write##size(ptr);                                                     \
 }
 
 DEFINE_TSAN_VOLATILE_READ_WRITE(1, 8);
@@ -411,7 +426,7 @@ DEFINE_TSAN_VOLATILE_READ_WRITE(8, 64);
    are needed because, without them, building a C library for OCaml with TSan
    enabled will fail at the linking step with an unresolved symbol error if it
    contains volatile accesses to 128-bit values. It is better to have 128-bit
-   volatiles behave silently like plain 128-bit values. */
+   volatiles behave silently like plain, non-volatile, 128-bit values. */
 
 extern void __tsan_read16(void*);
 extern void __tsan_write16(void*);
@@ -432,3 +447,17 @@ CAMLno_tsan void __tsan_unaligned_volatile_write16(void *ptr)
 {
   __tsan_write16(ptr);
 }
+
+CAMLno_tsan void caml_tsan_func_exit_asm(void) {
+#if defined(HAVE___TSAN_FUNC_EXIT_VOID_VOID_P)
+  __tsan_func_exit(NULL);
+#elif defined(HAVE___TSAN_FUNC_EXIT_VOID_VOID)
+  __tsan_func_exit();
+  #endif
+}
+
+CAMLno_tsan void caml_tsan_func_entry_asm(void *retaddr) {
+  __tsan_func_entry(retaddr);
+}
+
+// caml_tsan_write8 is never used in .S files
