@@ -25,8 +25,7 @@ Line 2, characters 6-23:
 2 |   fun (Either (Y a, N)) -> a;;
           ^^^^^^^^^^^^^^^^^
 Warning 8 [partial-match]: this pattern-matching is not exhaustive.
-Here is an example of a case that is not matched:
-Either (N, Y _)
+  Here is an example of a case that is not matched: "Either (N, Y _)"
 
 val f : filled either -> string = <fun>
 |}]
